

# ------------------------------------------------------------------------------
if __name__ == '__main__':
    import sys
    import qfracture

    # -- The first argument is always the filepath being executed
    # -- by python, so ignore that.
    all_args = sys.argv[1:]

    # -- Define our args and keyword arguments. We need to seperate these
    args = list()
    kwargs = dict()

    for arg in all_args:

        # -- If it has an =, its a kwarg
        if '=' in arg:
            kwargs[arg.split('=')[0]] = arg.split('=')[1]

        else:
            # -- We're dealing with an arg, so just
            # -- add it
            args.append(arg)

    # -- Launch the panel window with the given arguments
    qfracture.launch(
        *args,
        **kwargs
    )



"""
qfracture is a Qt based front end exposing a fracture project. The interface
allows for data browsing and searching as well as data interaction.
"""
from .core import launch
from .core import FractureUi
from .core import ElementDelegate

from . import tray

from . import constants



import qute
import fracture
import qfracture

import os


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class DefaultElementDelegate(qfracture.ElementDelegate):
    """
    This delegate is what we use by default if no other delegate
    is able to represent a piece of data
    """
    # -- Plugin Requirements
    name = 'Default Delegate'

    representing = [

    ]

    # -- Fonts
    NAME_FONT = qute.QFont('Ariel', 8, qute.QFont.Normal)
    DESC_FONT = qute.QFont('Ariel', 8, qute.QFont.StyleItalic)
    LARGE_NAME_FONT = qute.QFont('Ariel', 10, qute.QFont.Normal)

    # -- Brushes
    NAME_PEN = qute.QPen(qute.QColor(255, 255, 255))
    ACTIVE_NAME_PEN = qute.QPen(qute.QColor(255, 255, 255))
    DESC_PEN = qute.QPen(qute.QColor(255, 255, 255, 150))

    # -- Static Values
    BORDER = 5

    DEFAULT_ICON = fracture.icon('fracture_65')

    # --------------------------------------------------------------------------
    def __init__(self, element, parent=None):
        super(DefaultElementDelegate, self).__init__(element, parent)

        self.name = self.element.label()
        self.identifier = self.element.identifier()
        self.icon = self.element.icon()

    # --------------------------------------------------------------------------
    def paint(self, painter, option, index):

        # -- Add extra details for the selected item
        background_opacity = 0
        if option.state & qute.QStyle.State_Selected:
            background_opacity = 25
        elif option.state & qute.QStyle.State_MouseOver:
            background_opacity = 15

        painter.fillRect(
            option.rect,
            qute.QColor(
                255,
                255,
                255,
                background_opacity,
            )
        )

        # -- Draw the main header
        painter.setFont(self.NAME_FONT)
        painter.setPen(self.NAME_PEN)

        # -- Draw the main entry name
        painter.drawText(
            option.rect.x() + self._SIZE_HINT + self.BORDER,
            option.rect.y() + 15,
            self.name,
        )

        # -- Define the descriptions parameters
        painter.setOpacity(0.5)
        painter.setFont(self.DESC_FONT)
        painter.setPen(self.DESC_PEN)

        # -- Draw the descriptive text
        painter.drawText(
            option.rect.x() + self._SIZE_HINT + self.BORDER,
            option.rect.y() + 30,
            self.identifier,
        )
        painter.setOpacity(1)

        # -- If we have an icon, use it (providing it exists)
        # -- otherwise we fall back to our default icon
        icon = self.icon
        if not icon or not os.path.exists(icon):
            icon = self.DEFAULT_ICON

        px = qute.QPixmap(icon).scaled(
            self._SIZE_HINT,
            self._SIZE_HINT,
            mode=qute.Qt.SmoothTransformation,
        )

        # -- Draw the icon
        painter.drawPixmap(
            qute.QRect(
                option.rect.x(),
                option.rect.y(),
                self._SIZE_HINT,
                self._SIZE_HINT,
            ),
            px,
        )

    # --------------------------------------------------------------------------
    def sizeHint(self, option, index):
        return qute.QSize(1, self._SIZE_HINT)



"""
This is the main module for the qFracture Interface
"""
import re
import os
import qute
import fracture
import scribble
import factories
import subprocess

from . import menus
from . import wizard
from . import browser
from . import constants


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
class FractureUi(qute.QWidget):
    """
    This is a generic Ui Element for interacting with 
    seed.
    """

    _TAG_SPLIT = re.compile('[, ]')

    # --------------------------------------------------------------------------
    def __init__(self,
                 project_path,
                 style='space',
                 environment_id='qfracture',
                 parent=None):
        super(FractureUi, self).__init__(parent=parent)
        
        # -- Store the environment id, this is what we use to 
        # -- store and apply settings with
        self.environment_id = environment_id
        
        # -- Setup our window properties, ensuring it has an icon etc
        self.setWindowIcon(qute.QIcon(fracture.icon('fracture_65_black')))
        self.setWindowTitle('Fracture')

        # -- Apply a css stylesheet to our ui to make it a little
        # -- less bland!
        qute.applyStyle(
            [
                style,
                constants.UI_STYLE,
            ],
            self,
        )

        # -- Define the delegate accessors once the seed it set
        self.delegates = None
        self._potential_delegates = list()

        # -- Get the default display delegate from the factory
        # -- so we can utilise it if no other delegates are matched
        self._default_delegate = None

        # -- Load in our ui file
        self.ui = qute.loadUi(constants.UI_FILE)

        # -- Add the ui to the layout of our widget
        self.setLayout(qute.slimify(qute.QVBoxLayout()))
        self.layout().addWidget(self.ui)

        # -- Define the variable where we will hold the currently
        # -- active fracture instance
        self.project = None

        # -- Load in our settings
        settings = scribble.get(self.environment_id)

        # -- Set the seed to the current item - this ensures
        # -- the ui is all up to date
        self.setProject(project_path or settings.get('last_loaded', None))

        # -- Set the window geometry if we have the settings
        self.window().setGeometry(
            *settings.get(
                'geometry',
                [
                    300,
                    300,
                    400,
                    400,
                ]
            )
        )

        # -- Hook up our signals and slots to ensure the interaction
        # -- of the UI works as the user would expect
        self.ui.tags.returnPressed.connect(self.query)
        self.ui.tags.textChanged.connect(self.query)
        self.ui.browseList.itemRequiresDelegate.connect(self.assignDelegate)

        self.ui.addScanLocation.clicked.connect(self.addScanLocation)
        self.ui.removeScanLocation.clicked.connect(self.removeScanLocation)

        self.ui.addPluginLocation.clicked.connect(self.addPluginLocation)
        self.ui.removePluginLocation.clicked.connect(self.removePluginLocation)

        self.ui.addIgnorePattern.clicked.connect(self.addIgnorePattern)
        self.ui.removeIgnorePattern.clicked.connect(self.removeIgnorePattern)

        self.ui.setupProject.clicked.connect(self.initiateWizard)
        self.ui.loadProject.clicked.connect(self.loadProject)
        self.ui.launchSystemTray.clicked.connect(self.launchSystemTray)

        # -- Hook up the context menu rules
        self.ui.results.setContextMenuPolicy(qute.Qt.CustomContextMenu)
        self.ui.results.customContextMenuRequested.connect(
            self.itemContext,
        )

        # -- Initiate a starting query - which will display
        # -- any items marked as favourites
        self.query()

    # --------------------------------------------------------------------------
    def launchSystemTray(self):
        """
        This will launch a standalone system tray scanner for the currently
        active project.

        :return:
        """
        subprocess.Popen(
            [
                'python',
                os.path.join(
                    os.path.dirname(__file__),
                    'tray',
                    'core.py',
                ),
                self.project.identifier,
            ],
        )

    # --------------------------------------------------------------------------
    def loadProject(self):
        """
        Prompts the user for a fracture project file and then loads
        that project.

        :return:
        """
        load_path, _ = qute.QFileDialog.getOpenFileName(
            self,
            'Select Fracture Project',
            '',
            'fracture (*.fracture)'
        )

        if load_path:
            self.setProject(load_path)

    # --------------------------------------------------------------------------
    def getDelegate(self, element):
        """
        Private method for returning the delegate to be used by 
        the given DataElement
        
        :param element: DataElement to be represented
        :type element: DataElement
        
        :return: ElementDelegate
        """
        for delegate in self._potential_delegates:
            for component in element.components():
                if str(component) in delegate.representing:
                    return delegate

        return self._default_delegate

    # --------------------------------------------------------------------------
    # noinspection PyUnusedLocal
    def query(self, *args, **kwargs):
        """
        This will initiate a query, populating the results list based
        on the tags given in the tag line edit.
         
        :return: 
        """
        if not self.project:
            return

        # -- Run a query from Seed, giving all the tags in the
        # -- tag line edit
        results = self.project.find(
            [
                tag
                for tag in self._TAG_SPLIT.split(self.ui.tags.text())
                if tag
            ] or ['$(fav)'],
            limit=self.ui.searchLimit.value()
        )

        # -- Clear the current view before we populate it
        # -- over again
        self.ui.results.clear()

        # -- Start cycling over our results
        for result in results:
            result = self.project.get(result)
            if not result:
                continue

            # -- Create a widget item and store our data element
            # -- within it
            item = qute.QListWidgetItem()

            # -- Setup the tooltip
            item.setToolTip(
                'Name: %s\nType: [%s]\n Path: %s' % (
                    result.label(),
                    str(result),
                    result.identifier(),
                ),
            )

            # -- Add the result to the list
            self.ui.results.addItem(item)

            item.element = result

            self.assignDelegate(item, self.ui.results)

    # --------------------------------------------------------------------------
    def assignDelegate(self, item, list_widget):
        """
        Assigns the delegeate to the item based on what components are
        applied to it.

        :param item: QListWidget item (with a DataElement instance bound to
            its .element property).

        :return: None
        """
        # -- Assign the delegate to the item
        list_widget.setItemDelegateForRow(
            list_widget.row(item),
            self.getDelegate(item.element)(item.element, parent=list_widget),
        )

    # --------------------------------------------------------------------------
    def setProject(self, project_path):
        """
        This will set the seed instance for the ui. A new Seed instance
        will be generated, and providing the identifier loads successfully
        the ui will be updated to reflect the change.
        
        :param project_path: Data Source identifier
        :type project_path: str
        
        :return: None 
        """
        if not project_path or not os.path.exists(project_path):
            return

        # -- Re-assign the project property
        self.project = fracture.load(project_path)

        # -- Clear our current project specific data ui elements
        self.ui.scanLocations.clear()
        self.ui.ignorePatterns.clear()
        self.ui.pluginLocations.clear()

        # -- Our plugin locations may be None, in which case
        # -- switch them to an empty list
        plugin_locations = self.project.plugin_locations() or list()

        # -- Populate our data ui elements
        for location in self.project.plugin_locations():
            self.ui.pluginLocations.addItem(location)

        for location in self.project.scan_locations():
            self.ui.scanLocations.addItem(location)

        for pattern in self.project.skip_regexes():
            self.ui.ignorePatterns.addItem(pattern)

        # -- Create a factory to hold all our delegate plugins. Delegates
        # -- allow per-data types to be drawn differently.
        self.delegates = factories.Factory(
            abstract=ElementDelegate,
            plugin_identifier='name',
            versioning_identifier='version',
            paths=plugin_locations + [constants.BUILTIN_PLUGIN_DIR],
        )
        
        # -- Add our bespoke delegates to the factory
        self.delegates.add_path(
            os.path.join(
                os.path.dirname(__file__),
                'delegates',
            ),
        )

        # -- As a matter of optimisation, we sort these plugins
        # -- by priority only once.
        self._potential_delegates = sorted(
            self.delegates.plugins(),
            key=lambda p: p.priority,
        )

        self._default_delegate = self.delegates.request('Default Delegate')

        # -- Set the project for our browse list
        self.ui.browseList.setProject(self.project)
        self.ui.browseList.populate()

        # -- Update the window title
        self.window().setWindowTitle(
            'Seed : %s' % os.path.basename(project_path).split('.')[0],
        )
        # -- Now we store the fact that this is the last loaded
        # -- project
        settings = scribble.get(self.environment_id)
        settings['last_loaded'] = project_path
        settings.save()

        # -- Log the switch
        fracture.log.info('Switched Project : %s' % project_path)

    # --------------------------------------------------------------------------
    def addScanLocation(self):
        """
        This will prompt for a new scan location, and providing the user
        hits ok, the location will be added to seed.
        
        :return: None 
        """
        # -- Prompt for the search location
        location, ok = qute.QInputDialog.getText(
            None,
            'Add Scan Location',
            'Enter location for scanning'
        )

        if not ok:
            return

        self.project.add_scan_location(location)
        self.ui.scanLocations.addItem(location)

    # --------------------------------------------------------------------------
    def removeScanLocation(self):
        """
        This will remove the currently selected scan location from the 
        seed instance.
        
        :return: None 
        """
        # -- Prompt for the search location
        result = qute.QMessageBox.question(
                self,
                'Remove Scan Location',
                'Are you sure you want to remove the scan location : %s' % (
                    self.ui.scanLocations.currentText()
                ),
                qute.QMessageBox.Yes,
                qute.QMessageBox.No,
            )

        if result != qute.QMessageBox.Yes:
            return

        self.project.remove_scan_location(
            self.ui.scanLocations.currentText(),
        )
        self.ui.scanLocations.removeItem(self.ui.scanLocations.currentIndex())

    # --------------------------------------------------------------------------
    def addPluginLocation(self):
        """
        This will prompt for a new scan location, and providing the user
        hits ok, the location will be added to seed.

        :return: None
        """
        # -- Sessions are given newest to oldest
        location = qute.QFileDialog.getExistingDirectory(
            self,
            'Pluigin Location',
        )

        if not location:
            return

        location = location.replace('//', '/')

        self.project.add_plugin_location(location)
        self.ui.pluginLocations.addItem(location)

    # --------------------------------------------------------------------------
    def removePluginLocation(self):
        """
        This will remove the currently selected scan location from the
        seed instance.

        :return: None
        """
        # -- Prompt for the search location
        result = qute.QMessageBox.question(
            self,
            'Remove Plugin Location',
            'Are you sure you want to remove the plugin location : %s' % (
                self.ui.pluginLocations.currentText()
            ),
            qute.QMessageBox.Yes,
            qute.QMessageBox.No,
        )

        if result != qute.QMessageBox.Yes:
            return

        self.project.remove_plugin_location(
            self.ui.pluginLocations.currentText(),
        )
        self.ui.pluginLocations.removeItem(self.ui.pluginLocations.currentIndex())

    # --------------------------------------------------------------------------
    def addIgnorePattern(self):
        """
        This will prompt the user for a new ignore pattern, and providing
        the user hits ok, the patter will be added to the seed instance.
        
        :return: None 
        """
        # -- Prompt for the search location
        pattern, ok = qute.QInputDialog.getText(
            None,
            'Add Ignore Pattern',
            'Enter pattern to add'
        )

        if not ok:
            return

        self.project.add_skip_regex(pattern)
        self.ui.ignorePatterns.addItem(pattern)

    # --------------------------------------------------------------------------
    def removeIgnorePattern(self):
        """
        This will remove the currently selected ignore pattern and remove
        it from the seed instance.
        
        :return: None 
        """
        # -- Prompt for the search location
        result = qute.QMessageBox.question(
                self,
                'Remove Ignore Pattern',
                'Are you sure you want to remove the pattern : %s' % (
                    self.ui.ignorePatterns.currentText()
                ),
                qute.QMessageBox.Yes,
                qute.QMessageBox.No,
            )

        if result != qute.QMessageBox.Yes:
            return

        self.project.remove_skip_regex(
            self.ui.ignorePatterns.currentText(),
        )
        self.ui.ignorePatterns.removeItem(self.ui.ignorePatterns.currentIndex())

    # --------------------------------------------------------------------------
    def itemContext(self, pos):
        """
        This will inspect whether there is an item under the mouse, and if
        one is found it will be queried for any actions, and all seed actions
        will be bound to it and a menu generated. 
        
        This is the function which exposes actions to the user.
        
        :param pos: QPoint
         
        :return: None 
        """
        # -- Get the item at the event position
        item = self.ui.results.itemAt(pos)

        # -- If there is no item at the mouse position
        # -- then we simply return
        if not item:
            return

        # -- Create the menu
        menu = menus.ElementMenu(
            element=item.element,
            parent=self
        )

        # -- Popup the menu
        menu.exec_(qute.QCursor().pos())

    # --------------------------------------------------------------------------
    # noinspection PyUnusedLocal
    def resizeEvent(self, event):
        """
        Switch between North and west as the user resizes the panel.
        """

        # -- Store the current window size in our scribble settings
        window = self.window()
        settings = scribble.get(self.environment_id)
        settings['geometry'] = [
            window.pos().x() + 7,
            window.pos().y() + 32,
            window.width(),
            window.height(),
        ]
        settings.save()

    # --------------------------------------------------------------------------
    def initiateWizard(self):
        """
        The wizard gives a guide interface to creating a new fracture project
        """
        # -- Instance the wizard and block
        wizard_window = wizard.ClassWizard()
        wizard_window.exec_()

        # -- Now that we're done, get the save location
        new_project_path = wizard_window.field('saveLocation')

        # -- Providing the save location is valid, lets set the project
        if new_project_path:
            self.setProject(new_project_path)


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
class ElementDelegate(qute.QStyledItemDelegate):
    """
    This is the main Ui Delegate abstract. All delegates must
    inherit from this.
    """
    name = 'AbstractDelegate'

    # -- Version attribute is used for overriding
    # -- a delegate
    version = 1

    # -- Priority is used to determine importance, as the
    # -- ui will take the first matched item. The higher the
    # -- number the higher importance it has
    priority = 0

    # -- This allows you to specify a list of data types your
    # -- delegate can represent.
    # -- Note, if you leave this value as None on your re-implementation
    # -- then your delegate will override the default delegate.
    representing = [

    ]

    # -- Tracker for icon size
    _SIZE_HINT = 40

    # --------------------------------------------------------------------------
    def __init__(self, element, parent=None):
        super(ElementDelegate, self).__init__(parent)

        self.element = element

    # --------------------------------------------------------------------------
    # noinspection PyUnusedLocal
    def paint(self, painter, option, index):
        pass

    # --------------------------------------------------------------------------
    # noinspection PyUnusedLocal
    def sizeHint(self, option, index):
        return qute.QSize(1, self._SIZE_HINT)


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
def launch(project_path=None, blocking=True):

    q_app = qute.qApp()

    # -- Create a window and embed our widget into it
    widget = FractureUi(
        project_path=project_path,
    )
    window = qute.QMainWindow(parent=qute.mainWindow())

    # -- Update the geometry of the window to the last stored
    # -- geometry
    window.setGeometry(widget.geometry())
    window.setCentralWidget(widget)

    # -- Set the window properties
    title = 'Fracture : '

    if widget.project:
        title += os.path.basename(widget.project.identifier).split('.')[0]

    window.setWindowTitle(title)
    window.setWindowIcon(
        qute.QIcon(
            fracture.icon('fracture_65'),
        ),
    )

    # -- Show the ui, and if we're blocking call the exec_
    window.show()

    if blocking:
        q_app.exec_()


# ------------------------------------------------------------------------------
# -- Register our custom widgets which are used to override
# -- the widgets in the ui file
try:
    # noinspection PyUnresolvedReferences
    loader = qute.QtUiTools.QUiLoader()
    loader.registerCustomWidget(browser.FractureListWidget)

except Exception:
    pass



"""
This is a plugin creator wizard to make it easy for users to generate
their own plugins without coding.
"""
import os
import sys
import qute
import fracture
import functools


# ------------------------------------------------------------------------------
def _get_resource(name):
    """
    This is a convinience function to get files from the resources directory
    and correct handle the slashing.

    :param name: Name of file to pull from the resource directory

    :return: Absolute path to the resource requested.
    """
    return os.path.join(
        os.path.dirname(__file__),
        '_res',
        name,
    ).replace('//', '/')


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
class ClassWizard(qute.QWizard):
    """
    This is the main wizard class which guides the user through the 
    plugin creation. 
    """

    # --------------------------------------------------------------------------
    def __init__(self, parent=None):
        super(ClassWizard, self).__init__(parent)

        self.setWizardStyle(self.ModernStyle)
        self.setWindowTitle('Project Setup')
        self.setWindowIcon(
            qute.QIcon(fracture.icon('fracture_65'))
        )

        self.setPixmap(
            self.WatermarkPixmap,
            _get_resource('wizard.png'),
        )

        qute.applyStyle(
            [
                'space',
                _get_resource('wizard.qss'),
            ],
            self,
        )

        # -- Add pages to the wizard
        self.addPage(DataRootsPage())
        self.addPage(PluginLocationsPage())

        self.savePage = SaveProjectPage()
        self.addPage(self.savePage)

    # --------------------------------------------------------------------------
    def updateScanProgress(self, info):
        """
        Triggered during a scan, and updates the ui
        """
        self.savePage.ui.scanOutput.setText(info)
        self.update()

    # --------------------------------------------------------------------------
    def accept(self):
        """
        Upon acceptence, we can generate a plugin from the template.
        """
        data_paths = self.field('exportDataPaths').split(';')
        plugin_paths = self.field('exportPluginPaths').split(';')
        save_location = self.field('saveLocation')

        project = fracture.create(
            project_path=save_location,
            plugin_locations=[path for path in plugin_paths if path]
        )

        # -- Tell the project where to look for data
        for location in data_paths:
            if location:
                project.add_scan_location(location)

        # -- Now we initiate a scan. This will cycle over all the
        # -- scan locations and scrape them for data
        project.scanned.connect(self.updateScanProgress)
        project.scan()

        super(ClassWizard, self).accept()


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming,PyMethodMayBeStatic
class BasePage(qute.QWizardPage):

    # --------------------------------------------------------------------------
    def __init__(self, parent=None):
        super(BasePage, self).__init__(parent=parent)

        self.setLayout(qute.slimify(qute.QVBoxLayout()))

    # --------------------------------------------------------------------------
    def addLocation(self, list_widget, export_widget):
        """
        Generic add location callback function which pops up a browser and
        places the data into the given export widget.
        """
        # -- Sessions are given newest to oldest
        folder_path = qute.QFileDialog.getExistingDirectory(
            self,
            'Data Search Root',
        )
        if not folder_path:
            return

        folder_path = folder_path.replace('//', '/')

        # -- Add the item
        list_widget.addItem(folder_path)
        self.serialiseData(list_widget, export_widget)

    # --------------------------------------------------------------------------
    def removeLocation(self, list_widget, export_widget):
        """
        Convenience function for removing the selected entry from a list
        widget and re-serialising the data.
        """
        list_widget.takeItem(list_widget.currentRow())
        self.serialiseData(list_widget, export_widget)

    # --------------------------------------------------------------------------
    def serialiseData(self, list_widget, export_widget):
        """
        Convenience function for writing out the data from a list widget
        into a line edit.
        """
        # -- Serialise the data
        export_data = ''
        for idx in range(list_widget.count()):
            export_data += '%s;' % list_widget.item(idx).text()
            export_widget.setText(export_data)


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class DataRootsPage(BasePage):

    # --------------------------------------------------------------------------
    def __init__(self, parent=None):
        super(DataRootsPage, self).__init__(parent)

        # -- Load in our ui element
        self.ui = qute.loadUi(_get_resource('wizard_page_one.ui'))
        self.layout().addWidget(self.ui)

        self.ui.exportDataPaths.setVisible(False)

        # -- Register the field
        self.registerField('exportDataPaths', self.ui.exportDataPaths)

        self.ui.addSearchRootButton.clicked.connect(
            functools.partial(
                self.addLocation,
                self.ui.searchRoots,
                self.ui.exportDataPaths,
            )
        )

        self.ui.removeSearchRootButton.clicked.connect(
            functools.partial(
                self.removeLocation,
                self.ui.searchRoots,
                self.ui.exportDataPaths,
            )
        )


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class PluginLocationsPage(BasePage):

    # --------------------------------------------------------------------------
    def __init__(self, parent=None):
        super(PluginLocationsPage, self).__init__(parent)

        # -- Load in our ui element
        self.ui = qute.loadUi(_get_resource('wizard_page_two.ui'))
        self.layout().addWidget(self.ui)

        self.ui.exportPluginPaths.setVisible(False)

        self.registerField('exportPluginPaths', self.ui.exportPluginPaths)

        self.ui.addPluginLocationButton.clicked.connect(
            functools.partial(
                self.addLocation,
                self.ui.pluginLocations,
                self.ui.exportPluginPaths,
            )
        )

        self.ui.removePluginLocationButton.clicked.connect(
            functools.partial(
                self.removeLocation,
                self.ui.pluginLocations,
                self.ui.exportPluginPaths,
            )
        )


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
class SaveProjectPage(BasePage):

    # --------------------------------------------------------------------------
    def __init__(self, parent=None):
        super(SaveProjectPage, self).__init__(parent)

        # -- Load in our ui element
        self.ui = qute.loadUi(_get_resource('wizard_page_three.ui'))
        self.layout().addWidget(self.ui)

        # -- Register the field
        self.registerField(
            'saveLocation',
            self.ui.saveLocation,
        )

        self.ui.setSaveLocationButton.clicked.connect(
            self.setSaveLocation,
        )

    # --------------------------------------------------------------------------
    def setSaveLocation(self):

        # -- Browse for a file
        path, _ = qute.QFileDialog.getSaveFileName(
            self,
            dir='',
            filter='fracture (*.fracture)'
        )
        if not path:
            return None

        self.ui.saveLocation.setText(path.replace('//', '/'))


# ------------------------------------------------------------------------------
def show():
    _wizard = ClassWizard()
    _wizard.exec_()


# ------------------------------------------------------------------------------
if __name__ == '__main__':
    app = qute.qApp()
    wizard = ClassWizard()
    wizard.show()
    sys.exit(app.exec_())



"""

"""
import os
import sys
import qute
import fracture
import functools


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
class FractureTray(qute.QSystemTrayIcon):

    def __init__(self,
                 project_path,
                 auto_scan=False,
                 interval=5,
                 *args,
                 **kwargs):
        super(FractureTray, self).__init__(*args, **kwargs)

        # -- Store the fracture identifier we're responsible for
        self.project_path = project_path
        self.auto_scan = auto_scan
        self.interval = interval

        # -- Store the fracture instance we will use for scanning
        self.project = fracture.load(project_path=project_path)

        # -- We will use this to hold a thread from
        # -- which we will carry out our scanning
        self.scan_thread = None

        # -- Define our context menu. We will update this whenever
        # -- the user requests the menu
        self.menu = qute.QMenu()
        self.setContextMenu(self.menu)

        # -- Set the icon of the tray item. We request
        # -- this resource from fracture directly
        self.setIcon(qute.QIcon(fracture.icon('fracture_65')))

        # -- Create the timer object which we will use
        # -- to define when a scan will occur
        self.timer = qute.QTimer()
        self.timer.setInterval(self.interval * 1000)

        # -- If we're set to auto scan, then start a scan
        # -- immediately
        if self.auto_scan:
            self.scan()
            self.timer.start()

        # -- Ensure that when the timer trickles down
        # -- we begin a scan
        self.timer.timeout.connect(self.scan)

        # -- Hook up signals and slots. We use this signal to generate
        # -- the menu on-demand. This way the data is always correct at
        # -- the point in time when the user requests it
        self.activated.connect(self.generateMenu)

    # --------------------------------------------------------------------------
    def scan(self):
        """
        This will trigger a scan providing there is not a scan currently
        in progress. 
        """
        if not self.scan_thread:

            # -- Create the new scan thread
            self.scan_thread = ScanThread(self.project)

            # -- Ensure it clears itself once its finished, and
            # -- initiate its start.
            self.scan_thread.finished.connect(self.clear_scan)
            self.scan_thread.start()

    # --------------------------------------------------------------------------
    def clear_scan(self):
        """
        This is called whenever a scan is complete. This performs any
        object clean up.
        """
        self.scan_thread = None

    # --------------------------------------------------------------------------
    def generateMenu(self):
        """
        This will update the contents of the menu to be reflective of the
        users current settings. 
        """
        # -- Clear the current menu
        self.menu.clear()

        # -- Add an item which is here to show the identifier as well
        # -- as the scan status
        action = qute.QAction(
            '%s [%s]' % (
                os.path.basename(self.project_path).split('.')[0],
                'Scanning' if self.scan_thread else 'Idle',
            ),
            self.menu,
        )
        self.menu.addAction(action)

        # -- Add our seperator
        self.menu.addSeparator()

        # -- Add actions to enable/disable auto scan
        tag = 'Disable' if self.auto_scan else 'Enable'

        action = qute.QAction(
            '%s Auto Scan' % tag,
            self.menu,
        )

        # -- Connect the signal event
        action.triggered.connect(
            functools.partial(
                self.set_auto_scan,
                not self.auto_scan
            )
        )

        # -- Add the action
        self.menu.addAction(action)

        # -- Add the time between scan item
        action = qute.QAction(
            'Set Interval (%s)' % self.interval,
            self.menu
        )

        action.triggered.connect(
            functools.partial(
                self.set_time_between_scan,
            )
        )

        # -- Add the item
        self.menu.addAction(action)

        # -- Add our seperator
        self.menu.addSeparator()

        action = qute.QAction(
            'Scan',
            self.menu,
        )
        action.triggered.connect(self.scan)
        self.menu.addAction(action)

        # -- Add our seperator
        self.menu.addSeparator()

        # -- Add our exit option
        action = qute.QAction('Exit', self.menu)
        action.triggered.connect(sys.exit)
        self.menu.addAction(action)

        # -- Regenerate the menu
        self.setContextMenu(self.menu)

    # --------------------------------------------------------------------------
    def set_auto_scan(self, value):
        """
        This will change the auto scan setting.
        
        :param value: The value to switch to 
        """
        self.auto_scan = value

        if self.auto_scan:
            self.timer.start()

        else:
            self.timer.stop()

    # --------------------------------------------------------------------------
    def set_time_between_scan(self):
        """
        This allows the user to tailor how long to run between 
        scans 
        """
        value, ok = qute.QInputDialog.getInt(
            None,
            'Time (in minutes) between scans',
            'This is minimum time between scans',
            self.interval,
            minValue=1,
            maxValue=10000,
            step=1,
        )

        if not ok:
            return

        # -- Update our interval variable and apply it to the
        # -- timer
        self.interval = value
        self.timer.stop()
        self.timer.setInterval(self.interval * 1000)

        if self.auto_scan:
            self.timer.start()


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class ScanThread(qute.QThread):
    """
    All our scanning is handled in a thread to prevent the rest
    of the ui elements from being blocked.
    """
    # --------------------------------------------------------------------------
    def __init__(self, fracture_project):
        super(ScanThread, self).__init__()
        self.fracture_project = fracture_project

    # --------------------------------------------------------------------------
    def run(self):
        self.fracture_project.scan()


# ------------------------------------------------------------------------------
def launch(fracture_project, auto_scan=False, interval=10):
    """
    This is a convinence entrypoint to start the system tray
    application.
    
    :param fracture_project: The absolute path to the seed project
    :param auto_scan: This decides whether it will auto scan by default
    :param interval: This defines the default time between scans 
    """
    q_app = qute.qApp([])
    q_app.setQuitOnLastWindowClosed(False)

    tray = FractureTray(
        fracture_project,
        auto_scan=auto_scan,
        interval=interval,
    )

    tray.show()

    q_app.exec_()


# ------------------------------------------------------------------------------
if __name__ == '__main__':
    launch(*sys.argv[1:])



"""
This holds a series of constant variables utilised across multiple
functions and modules.
"""
import os


# ------------------------------------------------------------------------------
# -- This is the location of our static resources
RESOURCES = os.path.join(
    os.path.dirname(__file__),
    '_res',
)


# ------------------------------------------------------------------------------
# -- This is the ui file (path) which describes the majority
# -- of the static interface
UI_FILE = os.path.join(
    RESOURCES,
    'fracture.ui',
)

UI_STYLE = os.path.join(
    RESOURCES,
    'qfracture.css',
)


# ------------------------------------------------------------------------------
# -- This is a fracture ui tag used to allow users to mark
# -- data elements as favourites
FAVOURITE_TAG = '$(fav)'


# ------------------------------------------------------------------------------
BUILTIN_PLUGIN_DIR = os.path.join(
    os.path.dirname(__file__),
    'delegates',
)


import os
import setuptools

short_description = 'qFracture is a lightweight UI atop of fracture'
if os.path.exists('README.md'):
    with open('README.md', 'r') as fh:
        long_description = fh.read()

else:
    long_description = short_description

setuptools.setup(
    name='qfracture',
    version='0.9.2',
    author='Mike Malinowski',
    author_email='mike@twisted.space',
    description=short_description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mikemalinowski/qfracture',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_data={
        '': ['_res/*.css', '_res/*.png', '_res/*.ui', 'delegates/*.py'],
    },
    install_requires=['factories', 'qute', 'scribble', 'fracture'],
    keywords="fracture data composite qfracture",
)



import qute
import collections

from . import constants


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class ElementMenu(qute.QMenu):

    # --------------------------------------------------------------------------
    def __init__(self, element, *args, **kwargs):
        super(ElementMenu, self).__init__(*args, **kwargs)

        self.element = element

        # -- Add the menu items from the plugins
        menu_dict = collections.OrderedDict()
        menu_dict.update(element.functionality())

        # -- Add the seed menu items
        menu_dict['-'] = None
        menu_dict['Copy Path'] = self.copyPath

        # -- Determine if we need to
        if constants.FAVOURITE_TAG in element.tags():
            menu_dict['Remove from Favourites'] = self.unmarkAsFavourite

        else:
            menu_dict['Add to Favourites'] = self.markAsFavourite

        # -- We list all the components of an element, so
        # -- build that structure up
        component_data = collections.OrderedDict()
        data_types = sorted(
            [
                component.data_type
                for component in element.components()
            ]
        )

        for data_type in data_types:
            component_data['[%s]' % data_type] = self.null

        # -- Expose it in a sub menu
        menu_dict['--'] = None
        menu_dict['Components'] = component_data

        # -- Use Qute to populate our menu based on the dictionary
        qute.menuFromDictionary(
            menu_dict,
            parent=self,
            icon_paths=[constants.RESOURCES]
        )

    # --------------------------------------------------------------------------
    def null(self):
        """
        A blank callable.
        """
        pass

    # --------------------------------------------------------------------------
    def copyPath(self):
        """
        Private function for exposing built-in functionality to the
        context menu.
        """
        clipboard = qute.QApplication.clipboard()
        clipboard.setText(self.element.identifier())

    # --------------------------------------------------------------------------
    def markAsFavourite(self):
        """
        Private function for exposing built-in functionality to the
        context menu.
        """
        self.element.tag(constants.FAVOURITE_TAG)

    # --------------------------------------------------------------------------
    def unmarkAsFavourite(self):
        """
        Private function for exposing built-in functionality to the
        context menu.
        """
        self.element.untag(constants.FAVOURITE_TAG)



"""
The browser aims to give a more guide exploration mechanism, where the
user can traverse up and down within search locations, finding DataElements
within.
"""
import os
import qute

from . import menus


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
class FractureListWidget(qute.QListWidget):
    """
    This is a specifically designed list widget which exposes navigational
    elements at the top of the list to allow for scan location traversal along
    with the exposure of DataElements at the bottom of the list which can be
    found within the given search root.

    It essentially forms a kind of file browser which is not restricted
    specifically to folder/files but instead it is fed from the scan plugins.

    This makes it possible to combine the results of things like folder queries
    as well as source control queries or http queries.
    """

    # -- This is the maximum number of entries we show any any
    # -- one time for upward navigation.
    MAX_ABOVE = 2
    
    itemRequiresDelegate = qute.Signal(object, object)
    
    # --------------------------------------------------------------------------
    def __init__(self, parent=None):
        super(FractureListWidget, self).__init__(parent=parent)

        # -- This will be the fracture project we're interacting with. This
        # -- is set through the setProject method.
        self.project = None

        # -- Track the current location as the user clicks through
        self._current_location = None

        # -- Ensure we update the ui whenever the user asks to
        # -- center on an item
        self.itemClicked.connect(self.centerOn)

        self.setContextMenuPolicy(qute.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(
            self.itemContext,
        )

    # --------------------------------------------------------------------------
    def setProject(self, project):
        """
        Sets the active fracture project, and gives reference to the parent
        widget - which is expected to be a FractureUi instance.

        :param project: fracture.Project

        :return:
        """
        self.project = project

    # --------------------------------------------------------------------------
    def centerOn(self, item):
        """
        Triggered when the user clicks on
        :param item:
        :return:
        """
        # -- Check whether this is a valid item to be centered on
        if not hasattr(item, 'identifier'):
            return

        # -- If there is no change, do nothing
        if self._current_location == item.identifier:
            return

        # -- Store the current location and ask the view to
        # -- repopulate
        self._current_location = item.identifier
        self.populate()

    # --------------------------------------------------------------------------
    def populate(self):
        """
        Updates the view, building the navigation elements first before
        drawing the results.

        :return: None
        """
        # -- Clear the current view
        self.clear()

        # -- If we do not have a current location then we just
        if not self._current_location:
            for root in self.project.scan_locations():
                self._addDelegatedItem(root, RootAccessDelegate)

        else:

            # -- Add an accessor to the root locations
            self._addDelegatedItem(None, RootAccessDelegate, '...top...')

            # -- Get a list of all the items above and below
            above_items, below_items = self.project.explore(
                self._current_location,
            )

            # -- Add them in reverse, so it makes top down sense
            for item in reversed(above_items[:self.MAX_ABOVE]):
                self._addDelegatedItem(item, AboveDelegate)

            # -- If we have a current location we add that in.
            self._addDelegatedItem(self._current_location, CurrentDelegate)

            # -- Now we need to add the items below
            for item in below_items:
                self._addDelegatedItem(item, BelowDelegate)

            # -- Get the immediate identifiers
            elements = self.project.scan(
                self._current_location,
                recursive=False,
                full=False
            )

            for item in elements:

                result = self.project.get(item)

                if not result:
                    continue

                # -- Create a widget item and store our data element
                # -- within it
                item = qute.QListWidgetItem()

                # -- Assign the element composite to the item
                item.element = result

                # -- Add the result to the list
                self.addItem(item)

                # -- Declare that this item is ready for any delegates
                # -- to be bound for it
                self.itemRequiresDelegate.emit(item, self)

    # --------------------------------------------------------------------------
    def itemContext(self, pos):
        """
        This will inspect whether there is an item under the mouse, and if
        one is found it will be queried for any actions, and all fractyre 
        actions will be bound to it and a menu generated.

        This is the function which exposes actions to the user.

        :param pos: QPoint

        :return: None
        """
        # -- Get the item at the event position
        item = self.itemAt(pos)

        # -- If there is no item at the mouse position
        # -- then we simply return
        if not item or not hasattr(item, 'element'):
            return

        # -- Create the menu
        menu = menus.ElementMenu(
            element=item.element,
            parent=self
        )

        # -- Popup the menu
        menu.exec_(qute.QCursor().pos())

    # --------------------------------------------------------------------------
    def _addDelegatedItem(self, identifier, delegate, label=None):
        """
        This is a convenience function for added a list widget item with a
        specified delegate, as this is a proccess done multiple times during
        the populate method.

        :param identifier: DataElement Identifier
        :param delegate: Delegate to assign
        :param label: Optional label to assign to the delegate

        :return: None
        """
        # -- Add an accessor to the root locations
        root_access_item = qute.QListWidgetItem('current')
        root_access_item.identifier = identifier

        # -- Add the result to the list
        self.addItem(root_access_item)

        # -- Assign the delegate to the item
        self.setItemDelegateForRow(
            self.row(root_access_item),
            delegate(label or identifier, parent=self),
        )


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences,PyPep8Naming
class DirectionalDelegate(qute.QStyledItemDelegate):
    """
    This delegate is what we use by default if no other delegate
    is able to represent a piece of data
    """

    # -- Fonts
    NAME_FONT = qute.QFont('Ariel', 8, qute.QFont.Normal)
    DESC_FONT = qute.QFont('Ariel', 8, qute.QFont.StyleItalic)
    LARGE_NAME_FONT = qute.QFont('Ariel', 10, qute.QFont.Normal)

    # -- Brushes
    BG_COLOR = [109, 214, 255]
    NAME_PEN = qute.QPen(qute.QColor(255, 255, 255))
    BLACK_PEN = qute.QPen(qute.QColor(0, 0, 0))
    ACTIVE_NAME_PEN = qute.QPen(qute.QColor(255, 255, 255))
    DESC_PEN = qute.QPen(qute.QColor(255, 255, 255, 150))

    # -- Static Values
    BORDER = 5

    SIZE_HINT = 25
    DEFAULT_ICON = os.path.join(
        os.path.dirname(__file__),
        '_res',
        'above.png',
    )

    # --------------------------------------------------------------------------
    def __init__(self, identifier, parent=None):
        super(DirectionalDelegate, self).__init__(parent)
        self.identifier = identifier

    # --------------------------------------------------------------------------
    # noinspection PyUnusedLocal
    def paint(self, painter, option, index):

        # -- Add extra details for the selected item
        background_opacity = 50
        if option.state & qute.QStyle.State_MouseOver:
            background_opacity = 100
        
        # -- Draw the background
        painter.fillRect(
            option.rect,
            qute.QColor(
                self.BG_COLOR[0],
                self.BG_COLOR[1],
                self.BG_COLOR[2],
                background_opacity,
            )
        )

        # -- Draw the main header
        painter.setFont(self.NAME_FONT)
        painter.setPen(self.NAME_PEN)

        # -- Draw the main entry name
        painter.drawText(
            option.rect.x() + self.SIZE_HINT + self.BORDER,
            option.rect.y() + 15,
            os.path.basename(self.identifier),
        )

        # -- If we have an icon, use it (providing it exists)
        # -- otherwise we fall back to our default icon
        # icon = self.icon
        # if not icon or not os.path.exists(icon):
        icon = self.DEFAULT_ICON

        px = qute.QPixmap(icon).scaled(
            self.SIZE_HINT,
            self.SIZE_HINT,
            mode=qute.Qt.SmoothTransformation,
        )

        # -- Draw the icon
        painter.drawPixmap(
            qute.QRect(
                option.rect.x(),
                option.rect.y(),
                self.SIZE_HINT,
                self.SIZE_HINT,
            ),
            px,
        )

        painter.setPen(self.BLACK_PEN)
        # -- Finally, draw the border
        painter.drawRect(
            option.rect,
        )

    # --------------------------------------------------------------------------
    # noinspection PyUnusedLocal
    def sizeHint(self, option, index):
        return qute.QSize(1, self.SIZE_HINT)


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class BelowDelegate(DirectionalDelegate):
    """
    This delegate is what we use by default if no other delegate
    is able to represent a piece of data
    """
    DEFAULT_ICON = os.path.join(
        os.path.dirname(__file__),
        '_res',
        'below.png',
    )

    BG_COLOR = [
        109,
        214,
        255,
    ]


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class AboveDelegate(DirectionalDelegate):
    """
    This delegate is what we use by default if no other delegate
    is able to represent a piece of data
    """
    DEFAULT_ICON = os.path.join(
        os.path.dirname(__file__),
        '_res',
        'above.png',
    )

    BG_COLOR = [
        50,
        50,
        50,
    ]


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class CurrentDelegate(DirectionalDelegate):
    """
    This delegate is what we use by default if no other delegate
    is able to represent a piece of data
    """
    DEFAULT_ICON = os.path.join(
        os.path.dirname(__file__),
        '_res',
        'current.png',
    )

    BG_COLOR = [
        100,
        255,
        100,
    ]


# ------------------------------------------------------------------------------
# noinspection PyUnresolvedReferences
class RootAccessDelegate(DirectionalDelegate):
    """
    This delegate is what we use by default if no other delegate
    is able to represent a piece of data
    """
    DEFAULT_ICON = os.path.join(
        os.path.dirname(__file__),
        '_res',
        'search.png',
    )

    BG_COLOR = [
        100,
        100,
        100,
    ]
